 ! Finds matrix inverse, matrix determinant, and vector cross product.

 MODULE matrix_operations

 USE precise

 IMPLICIT NONE
  
 CONTAINS
  
 function matrixinv(matrix,n) result(b)
 !===================================================================================
 ! Code modified from the subroutine to calculate the inverse of a matrix using Gauss
 ! -Jordan elimination the inverse of matrix a(n,n) is calculated and stored in the 
 ! matrix b(n,n), from:
 ! http://www.webpages.uidaho.edu/~gabrielp/ME549-CE546/matrix-inverse.pdf
 ! accessed on: June 09 2016, 4:17 p.m.
 !========================================================================
 integer :: i,j,k,l,m,n,irow
 real(kind=dp),intent(in)::matrix(n,n)
 real(kind=dp):: a(n,n),big,dum,zero_tolerance=1.0e-14_dp,adjointA(2,2)
 real(kind=dp)::b(n,n)
 
 a=matrix
 if (n==1) then
   if (abs(a(1,1)) < zero_tolerance) then
     b = 0.0_dp
   else
     b = 1.0_dp/a(1,1) 
   end if
   RETURN
 else if (n==2) then
   if ( abs(a(1,1)*a(2,2)-(a(1,2)*a(2,1))) < zero_tolerance ) then
     !The matrix is singular and it has no inverse
     b=0.0_dp
     RETURN
   end if
   adjointA(1,1:2) = [a(2,2),-a(1,2)]
   adjointA(2,1:2) = [-a(2,1),a(1,1)]
   b=(1.0_dp/(a(1,1)*a(2,2)-a(2,1)*a(1,2)))*adjointA
   RETURN
 else
   !build the identity matrix
   b=0.0_dp
   do i = 1,n
     b(i,i) = 1.0_dp
   end do
   do i = 1,n ! this is the big loop over all the columns of a(n,n)
     ! in case the entry a(i,i) is zero, we need to find a good pivot; this pivot 
     ! is chosen as the largest value on the column i from a(j,i) with j = 1,n
     big = abs(a(i,i))
     do j = i+1,n
       if (abs(a(j,i)) .gt. big) then
         big = abs(a(j,i))
         irow = j
       end if
     end do
     ! interchange lines i with irow for both a() and b() matrices
     if (big .gt. abs(a(i,i))) then
       do k = 1,n
         dum = a(i,k) ! matrix a()
         a(i,k) = a(irow,k)
         a(irow,k) = dum
         dum = b(i,k) ! matrix b()
         b(i,k) = b(irow,k)
         b(irow,k) = dum
       end do
     end if
     ! divide all entries in line i from a(i,j) by the value a(i,i); 
     ! same operation for the identity matrix
     dum = a(i,i)
     if (abs(dum)  < zero_tolerance) then
       b=0.0_dp
       RETURN
     else
       do j = 1,n
         a(i,j) = a(i,j)/dum
         b(i,j) = b(i,j)/dum
       end do
     end if
     ! make zero all entries in the column a(j,i); same operation for b(j,i)
     do j = i+1,n
       dum = a(j,i)
       do k = 1,n
         a(j,k) = a(j,k) - dum*a(i,k)
         b(j,k) = b(j,k) - dum*b(i,k)
       end do
     end do
   end do
   ! substract appropiate multiple of row j from row j-1 
   do i = 1,n-1
     do j = i+1,n
       dum = a(i,j)
       do l = 1,n
         a(i,l) = a(i,l)-dum*a(j,l)
         b(i,l) = b(i,l)-dum*b(j,l)
       end do
     end do
   end do
 end if
 end function
 
 
 FUNCTION determinant(matrixa, n)
 !===================================================================================
 !Function to find the determinant of a square matrix
 !Author : Louisda16th a.k.a Ashwith J. Rego
 !Description: The subroutine is based on two key points:
 !1] A determinant is unaltered when row operations are performed: Hence, using this principle,
 !row operations (column operations would work as well) are used
 !to convert the matrix into upper traingular form
 !2]The determinant of a triangular matrix is obtained by finding the product of the diagonal elements
 !
 ! http://web.hku.hk/~gdli/UsefulFiles/Example-Fortran-program.html
 ! accessed on: June 14 2016, 10:18 a.m.
 !========================================================================
 REAL(kind=dp), DIMENSION(n,n) :: matrixa,matrix
 INTEGER, INTENT(IN) :: n
 REAL(kind=dp) :: m,determinant,zero_tolerance=1.0e-14_dp,big,dum
 INTEGER :: i, j, k, l,irow
 
 matrix=matrixa
 IF (n==1) THEN
   determinant = matrix(1,1)
   RETURN
 ELSE IF (n==2) THEN
   determinant = matrix(1,1)*matrix(2,2)-(matrix(1,2)*matrix(2,1))
   RETURN
 ELSE IF (n==3) THEN
   determinant=matrix(1,1)*matrix(2,2)*matrix(3,3)+matrix(1,2)*matrix(2,3)*matrix(3,1)+matrix(1,3)*matrix(2,1)*&
   matrix(3,2)-matrix(1,3)*matrix(2,2)*matrix(3,1)-matrix(1,2)*matrix(2,1)*matrix(3,3)-matrix(1,1)*&
   matrix(2,3)*matrix(3,2)
 ELSE
   l = 1
   !Convert to upper triangular form
   DO k = 1, n-1
     big = abs(matrix(k,k))
     do j = k+1,n
       if (abs(matrix(j,k)) .gt. big) then
         big = abs(matrix(j,k))
         irow = j
       end if
     end do
     IF (big < zero_tolerance) THEN
         determinant= 0.0_dp
         RETURN
     END IF
     ! interchange lines
     if (big .gt. abs(matrix(k,k))) then
       do i = 1,n
         dum = matrix(k,i) ! matrix a()
         matrix(k,i) = matrix(irow,i)
         matrix(irow,i) = dum
       end do
       l=-l
     end if
     DO j = k+1, n
       m = matrix(j,k)/matrix(k,k)
       DO i = k+1, n
         matrix(j,i) = matrix(j,i) - m*matrix(k,i)
       END DO
     END DO
   END DO
   !Calculate determinant by finding product of diagonal elements
   determinant= l
   DO i = 1, n
     determinant= determinant* matrix(i,i)
   END DO
 END IF
    
 END FUNCTION determinant

 
 
 FUNCTION cross(a,b)
 !============================================================
 ! Cross Product of two vectors for REAL numbers
 !------------------------------------------------------------
 ! http://stackoverflow.com/questions/6511711/computing-the-cr
 ! oss-product-of-two-vectors-in-fortran-90
 ! February 11th 2013 
 !===========================================================
 REAL(kind=dp), DIMENSION(3) :: cross
 REAL(kind=dp), DIMENSION(3), INTENT(IN) :: a, b

 cross(1) = a(2) * b(3) - a(3) * b(2)
 cross(2) = a(3) * b(1) - a(1) * b(3)
 cross(3) = a(1) * b(2) - a(2) * b(1)

 END FUNCTION cross


 
 END MODULE matrix_operations